
/********************** Measure DC-current Solar system V 0.6 ******************************
 *  DC Solar Current measure from PA3BYB in according to Solarduino HTTPS://solarduino.com
 * ***************************************************************************************** 
 * 
 *  The aim of the program is to switch on a water boiler if the generated solar energy is
 *  sufficient for this.
 *  In order to measure the DC-current coming from the solar panels, non-contact mesuring is
 *  necessary. A HALL-sensor of the HSTS16L/20A type is used for this.
 *  
 *  ******************* How the program works **********************************************
 *  After initialization and setup routine, the program start with selecting the buttons.
 *  When button is pressed, the offsetRead selector is set (1). This means  that loop routine
 *  is continued with the -DC curren Measurement- routine. The result in "currentResult" is 
 *  then used in the -Offset calculate- routine. 
 *  The loop wil be now -Calculate DC-current --> Offset calculate --> for at least 2.5 sec.
 *  After 2.5 sec, the measured result is placed in "currentOffset" and the program go back
 *  to central part where it displayed the result.
 *  
 *  ******************* HALL-sensor HSTS16L/20A ********************************************
 *  HALL Sensor center voltage is 2,5V The deviation voltage arround the center voltage is 
 *  max +/- 0,625 V for 20 Amp full scale. This means that 1 Amp is 625/20 = 31.25 mV/Amp.
 *  This reasult is needed for the float mVperAmpValue.
 *  
 *  The HALL-sensor has 5 wires. To connect I used a DIN-connector 5 pole met pennen verdeeld
 *  over 240 degrees.
 *  pin 1 = Black        = 0 V
 *  pin 2 = White        = Vref
 *  pin 3 = Cable shield = Gnd
 *  pin 4 = Yellow       = Vout
 *  pin 5 = Red          = +5 V
 *  
 *  The HALL-sensor is magnetic sensitive and has a small deviation of +/- 625 mV. This means 
 *  that current result measured can affected by magnetic radiation. So the measured current 
 *  result will be not exact. Even when no measurement is made there can be a result!!
 *  Therefore a "NUL" offset is needed. In this program it will be done automatically. 
 *  In situation no current flows true the HALL-sensor a measure is made. The result (+ or -) 
 *  will be stored in currentOffset. In measure current the offset will be substract or count 
 *  by the measured result and then displayed. 
 *  
 *  ******************* Board type used ****************************************************
 *  The board type I used is a Nano compatible type with the ATmega328. 
 *  It is a older type (<2018) therefore the IDE must be set to it. This is done as follows,
 *  select:<<tools>> <<Processor>> <<ATmega328P>> with old <<Old Boot Loader>>
 *  
 *  The Nano can operate on three different voltage sources: 
 *  - 6 - 20 V unregulated pin 30; 
 *  - 5V regulated         pin 27;
 *  - USB
 *  
 *  The built-in ADC is sensitive to supply voltage variation. If the supply voltage varies 
 *  due to load, the ADC will also vary accordingly.
 *  With a voltage variation due to load 4.95 V - 5.05 V, which is 100mV, the ADC output also 
 *  fluctuates in the same order. If you try to measure a very small voltage of 625 mV then a 
 *  deviation of 100 mV is about 16% !!!! That is far too much for an accurate measurement.
 *  The only solution is either to connect an extremely stable power supply or split up the 
 *  powersupplay in one for the NANO and one for the load.
 *  
 *  A1 = Analog pin in
 *  The program code read the voltage on pin A1 of the Hall-sensor
 *  
 *  D2 = Digital out pin
 *  The program write to this pin to switch the relay
 *  
 ****************************** Relais *****************************************************
 * Relais module with optocoupler for voltage separation. Type 5V and one changeover contact.
 * Trigger current 5mA. Max load AC 250V/10A, DC 30v/10A
 * When trigger voltage is HIGH the relay is in rest.
 *  
 *  
 ***************************** Some Functions **********************************************
 * "float"
 * float is a datatype for floating-point numbers. Floating-point numbers are often used to 
 * approximate analog and continuous values because they have greater resolution than integers. 
 * Floating-point numbers can be as large as 3.4028235E+38 and as low as -3.4028235E+38. 
 * They are stored as 32 bits (4 bytes) of information.
 * 
 * "millis()"
 * millis() Returns the number of milliseconds passed since the Arduino board began running 
 * the current program. This number will overflow (go back to zero), after approximately 
 * 50 days.
 * 
 * "analogRead()"
 * analogRead() Reads the value from the specified analog pin. Arduino boards contain a 
 * multichannel, 10-bit analog to digital converter. This means that it will map input 
 * voltages between 0 and the operating voltage 5V into integer values between 0 and 1023. 
 * On an Arduino UNO this yields a resolution between readings of: 5 volts / 1024 units 
 * or, 0.0049 volts (4.9 mV) per unit.
 * 
 * "void()"
 * The "void()" keyword is used only in functions declaration. It indicates the the function 
 * is expected to return NO information to the function from wich it was called.
 * Use in program it works like a subroutine.
 * Insite a subroutine is is NOT possible to run a second (or more) sub-sub routines. Only 
 * one subroutine is possible.
 * Another problem with subroutines is that they take a lot of time. Sub-routines are slow 
 * down the main program very much.
 *  
 */

//////////////////////////// Initialization Fix constants //////////////////////////////////
//
         //******** Initialization  LCD Display 1602A 16x2 charachters ***********
// The Display use data pin 4, 5, 6, 7, and 8 = RS; 9 = E; RW is connect to GND

  #include<LiquidCrystal.h>                              // Load LCD Library
  LiquidCrystal lcd(8,9,4,5,6,7);                        // LCD-dislay 1602
  unsigned long startMillisLCD;
  unsigned long currentMillisLCD;
  const unsigned long periodLCD = 1000;
  int numRows = 2;                                       // Init number of rows Display
  int numCols = 16;                                      // Init number of character Display

         //******** Initialization  HALL-sensor HSTS16L20A **********************
// The HSTS16L/20A Hall-Effect Current Transformer rated at 20A, 2.5V +/- 0.625V, 
// This means that the mVperAmp will be 625 mV / 20A = 31.25 mV/A 
  float mVperAmpValue     = 31.25 ;                     // Denominator can't be "nul"!!!

        //******* Initialization Reading pin Arduino is A1 **********************
// The A-series pin's Arduino Uno are analog reading pin with a ADC function. This means the
// voltage will be convert in byte words; 2 for integer and 4 for floating point numbers
  int analogPin = A1;                            // ArduinoUno pin A1 is analog "in"
  int decimalPrecision = 2;                      // max two digids behind the floating point

        //******* Initialization Relais on Digital Pin 12 ***********************
  int digitalPin = 2 ;                          // Digital Pin 2 is used for switching relay



///////////////////////////// Initialization Variabele  ////////////////////////////////////
//
//
        //***** Initialization DC Current Offset variables **********************
int offsetRead          = 0;                // To switch functions autocalibrate/normal
float middleVoltage     = 2500;             // Center voltage from the HALL-sensor
float supplyVoltage     = 5060;             // This voltage depends on the supply voltage 
float currentOffset     = 0.00;             // The (auto) offset deviation from autocalibrate
float offsetLastSample  = 0;                // to count time for each sample (1 sample                                           // is 1 milli
float offsetSampleCount = 0;                // Count number of sample


        //******** Initialization DC current measurement ************************
float sampleRead    = 0;               // to read the value of a sample
float sampleLast    = 0;               // Count the time for each sample (1 milli 1 sample)
float sampleSum     = 0;               // Accumulation of sample readings
float sampleCount   = 0;               // To count the number of samples
float currentMean   = 0;               // result of calculate the average value
float roughResult   = 0;               // result without offset value
float currentResult = 0;               // final result of the measure



/**************************************************************************************
*                                START PROGRAM
***************************************************************************************/

void setup() {
  pinMode(digitalPin, OUTPUT);                       // Digital pin 12 is now output
  digitalWrite(digitalPin, HIGH);                    // Relay is switch off
  
  Serial.begin(9600);                                // SetUp serial display
  lcd.setCursor(0,0);                                // Cursor placed on first character
  lcd.begin(numCols,numRows);                        // LCD has 16 characters in 2 rows
  
  startMillisLCD = millis();                         // start counting time for Display

  lcd.print("   PA3BYB   ");
  lcd.setCursor(0,1);
  lcd.print("Solar Amp meter");
  delay(4000);
  lcd.clear();

  offsetRead = 1;                                    // Start witch offset routine
}


//********************************** Main Loop ****************************************

void loop() {  
  
               //************** Button Select ****************


  int buttonRead;                                    // Read pressed button
  buttonRead = analogRead(0);                        // Read analog pin A0
  if (buttonRead < 60)                               // 250 mV / 4.9 mV = 127 units
    { lcd.clear(); lcd.setCursor(0,0); lcd.print ("Press Offset"); delay(1000);}
  else if (buttonRead <200)                          // 980 mV / 4.9 mV = 706 units
    { lcd.clear(); lcd.setCursor(0,0); lcd.print ("Offset routine"); delay(1000);
      
      offsetRead = 1;                                // Start Offset sub-routine
    }

  dcCurrent();

  displayCurrent();

}
 
      
               //********* Subroutine DC Current Measurement **************
void dcCurrent()
{
  if(millis() >= sampleLast + 1);                                    // Every one milli sec. taken one reading
    {
    sampleRead  = analogRead(analogPin) - ((middleVoltage/supplyVoltage)*1024);   // Read sample
    sampleSum   = sampleSum + sampleRead;
    sampleCount = sampleCount + 1 ;                             // next count
    sampleLast  = millis();                                     // reset time for next sample read
    }
  if (sampleCount == 1000)                                       // after 1 sec.) do calculation and display result
    {
    currentMean = sampleSum/sampleCount;                        // calculate average current value
    roughResult = (((currentMean/1024)*supplyVoltage)/mVperAmpValue);    // calculate rough result in amp
    currentResult = roughResult + currentOffset;                // Final result is result + offset

    sampleSum   = 0;                                            // reset accumulate sample for next cycle
    sampleCount = 0;                                            // reset nummber of sample for next cycle
    }


               //**************** Offset calculate *****************

//  when offsetRead = 1 the "Offset calculate" program part will be run

  if (offsetRead == 1)                                         // calulate the offset current
    { 
    currentOffset = 0;
    if (millis() > offsetLastSample)                           // Every one milli sec. taken one reading
      {
        offsetSampleCount = offsetSampleCount + 1;
        offsetLastSample = millis();
      }
    if (offsetSampleCount == 2500)            // If offsetSampleCount # 2500 (2.5 sec) go back to DC current measure
      {                                       // 
        currentOffset = - currentResult;      // currentOffeset == currentResult
        
        offsetSampleCount = 0;                // Reset for next offest calculate
        offsetRead = 0;                       // Calculate offset is ready, go back to DC current measurement
      }
   }
}
  
              //********* Display Momentane Current **********
void displayCurrent()
{
  currentMillisLCD = millis();
  if (currentMillisLCD - startMillisLCD >= periodLCD)
     {
       lcd.clear();
       lcd.setCursor(0,1);
       lcd.print(currentResult,decimalPrecision);
       lcd.print(" Amp");
       (startMillisLCD = currentMillisLCD); 
     }   

  if (currentResult >= 2)
     { digitalWrite(digitalPin, LOW); lcd.setCursor(0,0); lcd.print ("Relay switch on");}
  if (currentResult <= 2)
     { digitalWrite(digitalPin, HIGH); lcd.setCursor(0,0); lcd.print ("Relay switch off");}
}


/******************************************************************************************
 * 

//************************* Switch Relais ************************************
    if (offsetRead == 0) {                    // if no offset button go further

      if (currentResult > 2)                   // if current is > 2 Amp
        {
          callSubRelay();                      // Switch relay
        }
    }
    

//****************************** Relay handling **********************************
void callSubRelay()
{
      digitalWrite(digitalPin, LOW);
      lcd.setCursor(0,0);
      lcd.print ("Relay switch on");
      delay (4000);
      digitalWrite(digitalPin, HIGH);
      lcd.setCursor(0,0);
      lcd.print ("Relay switch off");
      delay (4000);
}
*/
